/* vim: ts=4
   rdesktop: A Remote Desktop Protocol client.
   User interface services - nano-X
   Copyright (C) Matthew Chapman 1999-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <nano-X.h>
#include <mwtypes.h>
#include <time.h>
#include "rdesktop.h"
#include <stdlib.h>

#include <assert.h>


//#undef DEBUG
//#define DEBUG(args...) fprintf(stderr, args);


#ifdef DEBUG
#	 define ENTER { DEBUG("Entering: %s\n",__PRETTY_FUNCTION__); }
#	 define LEAVE { DEBUG("Leaving: %s\n",__PRETTY_FUNCTION__); }
#else
#	 define ENTER
#	 define LEAVE
#endif

#ifdef IPAQ
extern GR_WINDOW_ID baseWindow;
#define IPAQ_W 240
#define IPAQ_H 320
static int dx;
static int dy;
static int lastScancode=0;
static int keyDown=0;
static int step = 4;
#endif
extern int width;
extern int height;
extern BOOL motion;

#ifndef true
#define true 1
#endif

#ifndef false
#define false 0
#endif


static GR_WINDOW_ID wnd;
static GR_DRAW_ID bstore;
static GR_GC_ID gc;
static GR_REGION_ID region = 0;
static int window_created = false;
static uint32 *colmap;
static unsigned int depth;
static int lastBg = 0, lastFg = 0;

static void xwin_set_function(GR_GC_ID aGC, uint8 rop2);
void ui_update_screen(int x, int y, int cx, int cy);


typedef struct _NxCursor
{
	int width;
	int height;
	int hotx;
	int hoty;
	MWIMAGEBITS *data;
	MWIMAGEBITS *mask;
} NxCursor;

static void nanoXErrorCallback(GR_EVENT *ep)
{
}

static GR_FNCALLBACKEVENT ErrorCallback = nanoXErrorCallback;

BOOL ui_create_window(char *title)
{
	GR_SCREEN_INFO sInfo;
	
	GrOpen();
	GrSetErrorHandler(ErrorCallback);

	window_created = true;
#ifdef IPAQ
	if (baseWindow == 0) baseWindow = GR_ROOT_WINDOW_ID;

    wnd = GrNewWindow(baseWindow, 0, 0, width, height, 0,0,0);
	GrSelectEvents(baseWindow, GR_EVENT_MASK_ALL);
	dx = 0;
	dy = 0;
#else
	wnd = GrNewWindow(GR_ROOT_WINDOW_ID, 0, 0, width, height, 0, 0, 0);
#endif
    bstore = GrNewPixmap(width, height, 0);

	GrMapWindow(wnd);
	gc = GrNewGC();
	GrSetGCForeground(gc, 0xffffff);
	GrSelectEvents(wnd, GR_EVENT_MASK_ALL);
	GrGetScreenInfo(&sInfo);
	depth = (sInfo.bpp == 8) ? 8 : 32;

	return true;
}

void ui_destroy_window()
{
    GrDestroyWindow(bstore);
	GrDestroyWindow(wnd);
	GrDestroyGC(gc);
	GrClose();
}

static uint16 xwin_translate_key(unsigned long key)
{
	DEBUG("KEY(code=0x%lx)\n", key);

	if ((key <= 0x60))
		return (key - 8);

	switch (key) {
		case 0x6a:  /* insert */
			return 0x152;
		case 0x6b:  /* delete */
			return 0x153;
		case 0x6c:  /* Numpad Enter */
			return 0x1c;
		case 0x6d:  /* Right ctrl */
			return 0x11d;
		case 0x61:  /* home */
			return 0x147;
		case 0x67:  /* end */
			return 0x14f;
		case 0x63:  /* page up */
			return 0x149;
		case 0x69:  /* page down */
			return 0x151;
		case 0x62:  /* up arrow */
			return 0x148;
		case 0x64:  /* left arrow */
			return 0x014b;
		case 0x66:  /* right arrow */
			return 0x14d;
		case 0x68:  /* down arrow */
			return 0x150;
		case 0x70:  /* Numpad forward slash */
			return 0x135;
		case 0x71:  /* Right alt */
			return 0x138;
		case 0x73:  /* Windows key */
			DEBUG("CHECKPOINT\n");
	}
	
	return 0;
}

static uint16 xwin_translate_mouse(unsigned long button)
{
	switch (button) {
		case 4:  /* left */
			return MOUSE_FLAG_BUTTON1;
		case 2:  /* middle */
			return MOUSE_FLAG_BUTTON3;
		case 1:  /* right */
			return MOUSE_FLAG_BUTTON2;
	}
	
	return 0;
}

static int num_state = 0;
static int cap_state = 0;

void reset_numcap_state()
{
	num_state = 0;
	cap_state = 0;
}

void check_lock(int state, int keycode, uint32 ev_time)
{
	switch (state) {
		case 0:
			if ((cap_state == 1) || (num_state == 1) || (keycode == 77) || (keycode == 66)) {
				if (keycode == 77) {
					num_state = 1;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 69, 0);
				}
				else if (keycode == 66) {
					cap_state = 1;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 58, 0);
				} else if (num_state == 0) {
					cap_state = 0;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				} else if (cap_state == 0) {
					num_state = 0;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
				}
				else {
					num_state = 0;
					cap_state = 0;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				}
			}

			break;

		case 2:

			if ((cap_state == 0) || (keycode == 66)) {
				if ((cap_state == 0) && (keycode == 66)) {
					break;
				}
				if (keycode == 66) {
					cap_state = 0;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 58, 0);
				} else {
					cap_state = 1;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				}
			}
			else if ((num_state == 1) || (keycode == 77)) {
				if (keycode == 77) {
					num_state = 1;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 69, 0);
				} else {
					num_state = 0;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
				}
			}


			break;

		case 16:
			//This will enable the Num lock

			if ((num_state == 0) || (keycode == 77)) {
				if ((num_state == 0) && (keycode == 77)) {
					break;
				} else if (keycode == 77) {
					num_state = 0;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 69, 0);
				} else {
					num_state = 1;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
				}
			}
			else if ((cap_state == 1) || (keycode == 66)) {
				if (keycode == 66) {
					cap_state = 1;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 58, 0);
				} else {
					cap_state = 0;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				}
			}

			break;

		case 18:

			if ((cap_state == 0) || (num_state == 0) || (keycode == 77) || (keycode == 66)) {
				if (((num_state == 0) && (keycode == 77)) || ((cap_state == 0) && (keycode == 66))) {
					break;
				}
				if (keycode == 77) {
					num_state = 0;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 69, 0);
				} else if (keycode == 66) {
					cap_state = 0;
					rdp_send_input(ev_time, RDP_INPUT_SCANCODE, 0, 58, 0);
				} else if (num_state == 1) {
					cap_state = 1;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				} else if (cap_state == 1) {
					num_state = 1;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
				}
				else {
					num_state = 1;
					cap_state = 1;
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 69, 0);
					rdp_send_input(ev_time - 1 , RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 69, 0);
					rdp_send_input(ev_time - 1, RDP_INPUT_SCANCODE, 0, 58, 0);
					rdp_send_input(ev_time - 1 , RDP_INPUT_SCANCODE, KBD_FLAG_DOWN | KBD_FLAG_UP, 58, 0);
				}
			}
	}
}

#ifdef IPAQ
int checkScrolling(GR_EVENT *event)
{
	switch(event->keystroke.scancode) {
		case 134:
			dy -= step;
			break;
		case 135:
			dx += step;
			break;
		case 136:
			dx -= step;
			break;
		case 137:
			dy += step;
			break;
		default:
			return 0;
	}
    if (dx < 0) dx = 0;
    if (dy < 0) dy = 0;
    if (dx+240 > width) dx = width - 240;
    if (dy+296 > height) dy = height - 296;

	ui_update_screen(0,0,width,height);
	return 1;
}
#endif

void ui_process_events()
{
	GR_EVENT event;
	uint16 scancode;
	uint16 button;
	uint32 ev_time;

	if (window_created == false)
		return ;

	while (GrPeekEvent(&event)) {
		GrGetNextEvent(&event);
		ev_time = time(NULL);

		switch (event.type) {
			case GR_EVENT_TYPE_KEY_DOWN:
#ifdef IPAQ
				keyDown = 1;
				scancode = 0;
				lastScancode = event.keystroke.scancode;

				if (!checkScrolling(&event))
#endif
				scancode = xwin_translate_key(event.keystroke.scancode);

				if (scancode == 0)
					break;

				rdp_send_input(ev_time, RDP_INPUT_SCANCODE,
				               (scancode & 0x100) ? 0x100 : 0x00, scancode, 0);
				break;

			case GR_EVENT_TYPE_KEY_UP:
#ifdef IPAQ
				keyDown = 0;
				lastScancode = 0;
#endif
				scancode = xwin_translate_key(event.keystroke.scancode);

				if (scancode == 0)
					break;

				/* Check scancode then send appropriate key down event as third
				 * argument. */
				rdp_send_input(ev_time, RDP_INPUT_SCANCODE,
				               (scancode & 0x100) ? 0xc100 : 0xc000, scancode, 0);
				break;

			case GR_EVENT_TYPE_BUTTON_DOWN:
				button = xwin_translate_mouse(event.button.changebuttons);

				if (button == 0)
					break;

				rdp_send_input(ev_time, RDP_INPUT_MOUSE,
				               button | MOUSE_FLAG_DOWN,
#ifdef IPAQ
				               event.button.x+dx,
				               event.button.y+dy);
#else
				               event.button.x,
				               event.button.y);
#endif
				break;

			case GR_EVENT_TYPE_BUTTON_UP:
				button = xwin_translate_mouse(event.button.changebuttons);
				if (button == 0)
					break;

				rdp_send_input(ev_time, RDP_INPUT_MOUSE,
				               button,
				               event.button.x,
				               event.button.y);
				break;

			case GR_EVENT_TYPE_MOUSE_POSITION:
				rdp_send_input(ev_time, RDP_INPUT_MOUSE,
				               MOUSE_FLAG_MOVE,
				               event.mouse.x,
				               event.mouse.y);
				break;

			case GR_EVENT_TYPE_UPDATE: 
				xwin_set_function(gc, ROP2_COPY);
				ui_update_screen(0,0, width, height);
				break;

			case GR_EVENT_TYPE_CLOSE_REQ:
				rdp_disconnect();
				ui_destroy_window();
				break;

			default:
				/*DEBUG("Unhandled event: %d\n",event.type);*/
				break;
		}
	}
#ifdef IPAQ
	if (keyDown)
		GrInjectKeyboardEvent(wnd, 0, 0, lastScancode, 1);
#endif
}

void ui_move_pointer(int x, int y)
{
	GrMoveCursor(x, y);
}

#define Ctrans(col) (((GR_PALETTE*)colmap)->palette[col])

static uint8 *translate8(int width, int height, uint8 *data)
{
	uint8 *d2 = xmalloc(width * height);
	memcpy(d2, data, width*height);
	return d2;
}

static uint8 *translate16(int width, int height, uint8 *data)
{
	uint32 i;
	uint8 *d2 = xmalloc(width * height * (depth / 8));
	uint8 *d3 = d2;

	for (i = width * height; i; i--) {
		GR_PALENTRY pix = Ctrans(*data++);
		*d3++ = pix.r;
		if ((depth / 8) > 1) {
			*d3++ = pix.g;
			if ((depth / 8) > 2) {
				*d3++ = pix.b;
				*d3++ = 0;
			}
		}
	}
	return d2;
}

static uint8 *translate(int width, int height, uint8 *data)
{
	return depth == 8 ? translate8(width, height, data) :
	       translate16(width, height, data);
}

HBITMAP ui_create_bitmap(int width, int height, uint8 *data)
{
	GR_DRAW_ID bitmap;
	uint8 *d2;

	d2 = translate(width, height, data);
	bitmap = GrNewPixmap(width, height, 0);
	
	GrArea(bitmap, gc, 0, 0, width, height, d2, MWPF_RGB);
	xfree(d2);

	return (HBITMAP)bitmap;
}

void ui_paint_bitmap(int x, int y, int cx, int cy,
                     int width, int height, uint8 *data)
{
	GR_DRAW_ID pix;
ENTER;

	pix = (GR_DRAW_ID)ui_create_bitmap(width, height, data);

    GrCopyArea(bstore, gc, x, y, cx, cy, pix, 0, 0, MWROP_COPY);

    ui_update_screen(x, y, cx, cy);
	ui_destroy_bitmap((HBITMAP)pix);
LEAVE;
}

void ui_destroy_bitmap(HBITMAP bmp)
{
	GrDestroyWindow((GR_DRAW_ID)bmp);
}

HCURSOR ui_create_cursor(unsigned int hot_x, unsigned int hot_y, int width, int height,
                         uint8 *mask, uint8 *data)
{
	return 0;
}

void ui_set_cursor(HCURSOR cursor)
{
/*
	NxCursor *c = (NxCursor*)cursor;
	GrSetCursor(wnd, 
				c->width,
				c->height,
				c->hotx,
				c->hoty,
				0xffffff,
				0,
				c->data,
				c->mask);
*/					
}

void ui_destroy_cursor(HCURSOR cursor)
{
	xfree((NxCursor*)cursor);
}

HGLYPH ui_create_glyph(int width, int height, uint8 *data)
{
	GR_IMAGE_HDR *bmp;
	uint8 d, d1, y;
	uint16 datasize;
ENTER;
	datasize = (height * ((width + 7) / 8) + 3) & ~3;

	bmp = (GR_IMAGE_HDR*)malloc(sizeof(GR_IMAGE_HDR));
	bmp->width = width;
	bmp->height = height;
	bmp->planes = 1;
	bmp->bpp = 1;
	bmp->pitch = (width + 7 ) / 8;
	bmp->bytesperpixel = 0;
	bmp->compression = MWIMAGE_RGB;
	bmp->palsize = 2;
	bmp->transcolor = 0;
	bmp->palette = (MWPALENTRY*)malloc(2 * sizeof(MWPALENTRY));
	bmp->palette[0].r = 0x0;
	bmp->palette[0].g = 0x0;
	bmp->palette[0].b = 0x0;
	bmp->palette[1].r = 0xff;
	bmp->palette[1].g = 0xff;
	bmp->palette[1].b = 0xff;
	bmp->imagebits = (MWUCHAR*)malloc(datasize);

	// Reverse the bits here LSB->MSB
	for (y = 0; y < datasize; y++) {
		d1 = data[y];
		d = 0;
		if (d1 & 1) d |= 128;
		if (d1 & 2) d |= 64;
		if (d1 & 4) d |= 32;
		if (d1 & 8) d |= 16;
		if (d1 & 16) d |= 8;
		if (d1 & 32) d |= 4;
		if (d1 & 64) d |= 2;
		if (d1 & 128) d |= 1;
		bmp->imagebits[y] = d;
	}
LEAVE;
	return (HGLYPH)bmp;
}

void ui_destroy_glyph(HGLYPH glyph)
{
ENTER;
	free((char*)((GR_IMAGE_HDR*)glyph)->imagebits);
	free((char*)((GR_IMAGE_HDR*)glyph)->palette);
	free((char*)glyph);
LEAVE;
}

HCOLOURMAP ui_create_colourmap16(COLOURMAP *colours)
{
	COLOURENTRY *entry;
	GR_PALETTE *pal;
	int i, ncolors = colours->ncolours;
ENTER;
	pal = (GR_PALETTE*)malloc(sizeof(GR_PALETTE));
	pal->count = ncolors;

	for (i = 0; i < ncolors; i++) {
		entry = &colours->colours[i];

		pal->palette[i].r = entry->red;
		pal->palette[i].g = entry->green;
		pal->palette[i].b = entry->blue;
	}
LEAVE;
	return (HCOLOURMAP)pal;
}

void ui_destroy_colourmap16(HCOLOURMAP map)
{
ENTER;
	free(map);
LEAVE;
}

void ui_set_colourmap16(HCOLOURMAP map)
{
ENTER;
	ui_destroy_colourmap(colmap);
	colmap = map;
LEAVE;
}

HCOLOURMAP ui_create_colourmap8(COLOURMAP *colours)
{
ENTER; LEAVE;
	return 0;
}

void ui_destroy_colourmap8(HCOLOURMAP map)
{
}

void ui_set_colourmap8(HCOLOURMAP map)
{
}

HCOLOURMAP ui_create_colourmap(COLOURMAP *colours)
{
	return depth == 8 ? ui_create_colourmap8(colours) :
	       ui_create_colourmap16(colours);
}

void ui_destroy_colourmap(HCOLOURMAP map)
{
	depth == 8 ? ui_destroy_colourmap8(map) : ui_destroy_colourmap16(map);
}

void ui_set_colourmap(HCOLOURMAP map)
{
	depth == 8 ? ui_set_colourmap8(map) : ui_set_colourmap16(map);
}

void ui_set_clip(int x, int y, int cx, int cy)
{
	GR_RECT rect;

	assert(x >= 0);
	assert(y >= 0);
	assert(cx >= 0);
	assert(cy >= 0);
	
	rect.x = x;
	rect.y = y;
	rect.width = cx;
	rect.height = cy;

	if (region != 0) 
		GrDestroyRegion(region);
	region = GrNewRegion();
	GrUnionRectWithRegion(region, &rect);
	GrSetGCRegion(gc, region);
}

void ui_reset_clip()
{
	GR_RECT rect;

	rect.x = 0;
	rect.y = 0;
	rect.width = width;
	rect.height = height;

	if (region != 0) 
		GrDestroyRegion(region);
	region = GrNewRegion();
	GrUnionRectWithRegion(region, &rect);
	GrSetGCRegion(gc, region);
}

void ui_bell()
{
	GrBell();
}

static int rop2_map[] = {
    GR_MODE_CLEAR,          /* 0 */
    GR_MODE_NOR,            /* DPon */
    GR_MODE_ANDINVERTED,    /* DPna */
    GR_MODE_COPYINVERTED,   /* Pn */
    GR_MODE_ANDREVERSE,     /* PDna */
    GR_MODE_INVERT,         /* Dn */
    GR_MODE_XOR,            /* DPx */
    GR_MODE_NAND,           /* DPan */
    GR_MODE_AND,            /* DPa */
    GR_MODE_EQUIV,          /* DPxn */
    GR_MODE_NOOP,           /* D */
    GR_MODE_ORINVERTED,     /* DPno */
    GR_MODE_COPY,           /* P */
    GR_MODE_ORREVERSE,      /* PDno */
    GR_MODE_OR,             /* DPo */
    GR_MODE_SETTO1          /* 1 */
};

static void xwin_set_function(GR_GC_ID aGC, uint8 rop2)
{
	GrSetGCMode(aGC, rop2_map[rop2]);
}

void ui_destblt(uint8 opcode,
                /* dest */  int x,  int y,  int cx,  int cy)
{
	GR_GC_ID aGC;

	aGC = GrCopyGC(gc);

	xwin_set_function(aGC, opcode);
/*
	GrSetGCForeground(aGC, lastFg);
	GrSetGCBackground(aGC, lastBg);

    GrFillRect(bstore, aGC, x, y, cx, cy);
*/
    ui_update_screen(x, y, cx, cy);

	GrDestroyGC(aGC);
}

void ui_patblt(uint8 opcode,
               /* dest */  int x,   int y,   int cx,   int cy,
               /* brush */ BRUSH *brush,   int bgcolour,   int fgcolour)
{
	GR_PALENTRY pal;
	GR_IMAGE_HDR *fill;
	GR_DRAW_ID pix;
	int fg, bg;
	int x1, y1;
	GR_GC_ID aGC;

ENTER;
	aGC = GrCopyGC(gc);
	xwin_set_function(aGC, opcode);

	pal = Ctrans(fgcolour);
	fg = MWRGB(pal.r, pal.g, pal.b);

	pal = Ctrans(bgcolour);
	bg = MWRGB(pal.r, pal.g, pal.b);

	GrSetGCForeground(aGC, fg);
	GrSetGCBackground(aGC, bg);

	lastBg = bg;
	lastFg = fg;

	switch(brush->style) {
		case 0: /* Solid */
//			printf("%s - SOLID\n",__FUNCTION__);
			GrSetGCUseBackground(aGC, false);
			GrFillRect(bstore, aGC, x, y, cx, cy);
			GrSetGCUseBackground(aGC, true);
		break;
		case 3: /* Pattern */
//			printf("%s - PATTERN\n",__FUNCTION__);

//FIXME: temporary hack to disable rubber bands for the iPAQ

#ifdef IPAQ
			if ((cx < 2) || (cy < 2)) break;
#endif
			
			GrSetGCUseBackground(aGC, false);

			fill = (GR_IMAGE_HDR*)ui_create_glyph(8, 8, brush->pattern);

			pix = GrNewPixmap(cx, cy, 0);

			GrFillRect(pix, aGC, 0,0, cx, cy);

			fill->palette[1] = Ctrans(bgcolour);
			fill->palette[0] = Ctrans(fgcolour);

			/* printf("%d: %d %d %d %d\n",opcode,x,y,cx,cy); */

			/* put pattern into all available areas of pix */
			for(x1 = 0; x1 < cx; x1 += fill->width)
				for(y1 = 0; y1 < cy; y1 += fill->height)
					GrDrawImageBits(pix, aGC, x1, y1, fill);

			GrCopyArea(bstore, aGC, x, y, cx, cy, pix, 0, 0, MWROP_COPY);
			
			ui_destroy_glyph((HGLYPH)fill);
			GrDestroyWindow(pix);
		break;
		default:
			NOTIMP("brush %d\n", brush->style);
	}
    ui_update_screen(x, y, cx, cy);
	GrDestroyGC(aGC);
LEAVE;
}

void ui_screenblt(uint8 opcode,
                  /* dest */ int x,  int y,  int cx,  int cy,
                  /* src */  int srcx,  int srcy)
{
	GR_GC_ID aGC;
	GR_DRAW_ID pixmap;
ENTER;
	
	aGC = GrCopyGC(gc);
	
	xwin_set_function(aGC, opcode);

	pixmap = GrNewPixmap(cx, cy, NULL);

	GrCopyArea(pixmap, aGC, 0, 0, cx, cy, 
				bstore, srcx, srcy, MWROP_COPY);

	GrCopyArea(bstore, aGC, x, y, cx, cy,
				pixmap, 0, 0, MWROP_COPY);

    ui_update_screen(x, y, cx, cy);
	GrDestroyWindow(pixmap);

	GrDestroyGC(aGC);
LEAVE;
}

void ui_memblt(uint8 opcode,
               /* dest */  int x,  int y,  int cx,  int cy,
               /* src */   HBITMAP src,  int srcx,  int srcy)
{
	GR_GC_ID aGC;
ENTER;
	
	aGC = GrCopyGC(gc);
	xwin_set_function(aGC, opcode);

	GrCopyArea(bstore, aGC, x, y, cx, cy,
	           (GR_DRAW_ID)src, srcx, srcy, MWROP_COPY);

    ui_update_screen(x, y, cx, cy);
	GrDestroyGC(aGC);
LEAVE;
}

void ui_triblt(uint8 opcode,
               /* dest */  int x,  int y,  int cx,  int cy,
               /* src */   HBITMAP src,  int srcx,  int srcy,
               /* brush */ BRUSH *brush,  int bgcolour,  int fgcolour)
{
	/* This is potentially difficult to do in general. Until someone
	   comes up with a more efficient way of doing it I am using cases. */
ENTER;
	switch (opcode) {
		case 0x69:  /* PDSxxn */
			printf("Opcode: 0x%x\n",opcode);
			ui_memblt(ROP2_XOR, x, y, cx, cy, src, srcx, srcy);
			ui_patblt(ROP2_NXOR, x, y, cx, cy,
			          brush, bgcolour,
			          fgcolour);
			break;

		case 0xb8:  /* PSDPxax */
			ui_patblt(ROP2_XOR, x, y, cx, cy,
			          brush, bgcolour,
			          fgcolour);
			ui_memblt(ROP2_AND, x, y, cx, cy, src, srcx, srcy);
			ui_patblt(ROP2_XOR, x, y, cx, cy,
			          brush, bgcolour,
			          fgcolour);
			break;

		default:
			NOTIMP("triblt 1x%x\n", opcode);
			ui_memblt(ROP2_COPY, x, y, cx, cy, src, srcx, srcy);
	}
LEAVE;
}

void ui_line(uint8 opcode,
             /* dest */  int startx,  int starty,  int endx,  int endy,
             /* pen */   PEN *pen)
{
	GR_PALENTRY pal;
	GR_GC_ID aGC;

	aGC = GrCopyGC(gc);

	xwin_set_function(aGC, opcode);

	pal = Ctrans(pen->colour);
	GrSetGCForeground(aGC, MWRGB(pal.r, pal.g, pal.b));
	GrLine(bstore, aGC, startx, starty, endx, endy);

	GrDestroyGC(aGC);
}

void ui_rect(
    /* dest */  int x,  int y,  int cx,  int cy,
    /* brush */ int colour)
{
	GR_PALENTRY pal;
	GR_GC_ID aGC;

	aGC = GrCopyGC(gc);

	xwin_set_function(aGC, ROP2_COPY);
	pal = Ctrans(colour);
	GrSetGCForeground(aGC, MWRGB(pal.r, pal.g, pal.b));

	GrFillRect(bstore, aGC, x, y, cx, cy);
    ui_update_screen(x, y, cx, cy);

	GrDestroyGC(aGC);
}

void ui_draw_glyph(int mixmode,
		/* dest */	int x,  int y,  int cx,  int cy,
		/* src */	HGLYPH glyph,  int srcx,  int srcy, 
					int bgcolour,  int fgcolour, HBITMAP dest)
{
	GR_PALENTRY savefg, savebg, pal;
	int fg, bg;
	GR_GC_ID aGC;

	aGC = GrCopyGC(gc);
	xwin_set_function(aGC, ROP2_COPY);

	pal = Ctrans(fgcolour);

	savefg = ((GR_IMAGE_HDR*)glyph)->palette[1];
	((GR_IMAGE_HDR*)glyph)->palette[1] = Ctrans(fgcolour);

	savebg = ((GR_IMAGE_HDR*)glyph)->palette[0];
	((GR_IMAGE_HDR*)glyph)->palette[0] = Ctrans(bgcolour);

	pal = Ctrans(fgcolour);
	fg = MWRGB(pal.r, pal.g, pal.b);
	GrSetGCForeground(aGC, fg);

	switch(mixmode) {
		case MIX_TRANSPARENT:
			GrSetGCUseBackground(aGC, false);
			break;
		case MIX_OPAQUE:
			pal = Ctrans(bgcolour);
			bg = MWRGB(pal.r, pal.g, pal.b);
			GrSetGCBackground(aGC, bg);
			break;
		default: 
			NOTIMP("mix: %d\n", mixmode);
	}

	GrDrawImageBits(bstore, aGC, x, y, (GR_IMAGE_HDR*)glyph);

	((GR_IMAGE_HDR*)glyph)->palette[1] = savefg;
	((GR_IMAGE_HDR*)glyph)->palette[0] = savebg;

	GrDestroyGC(aGC);
}

void ui_draw_text(uint8 font, uint8 flags, int mixmode, int x, int y,
                  int clipx, int clipy, int clipcx, int clipcy,
                  int boxx, int boxy, int boxcx, int boxcy,
                  int bgcolour, int fgcolour, uint8 *text, uint8 length)
{
	FONTGLYPH *glyph;
	int i;
	int xoffset;
	GR_PALENTRY pal;

	pal = Ctrans(fgcolour);

	if (boxcx > 1) {
		ui_rect(boxx, boxy, boxcx, boxcy, bgcolour);
	} else if (mixmode == MIX_OPAQUE) {
		ui_rect(clipx, clipy, clipcx, clipcy, bgcolour);
	}

	/* Paint text, character by character */
	for (i = 0; i < length; i++) {
		glyph = cache_get_font(font, text[i]);

        if (!(flags & TEXT2_IMPLICIT_X)) {
            xoffset = text[++i];
            if (xoffset & 0x80)
                /* offset is really the following 2 bytes
                 * (16-bit number, little-endian) */
                x += text[++i] | (text[++i] << 8);
            else
                x += xoffset;
        }       

		if (glyph != NULL) {
			ui_draw_glyph(mixmode, x + (short)glyph->offset,
			              y + (short)glyph->baseline,
			              glyph->width, glyph->height,
			              glyph->pixmap, 0, 0,
			              bgcolour, fgcolour, &bstore);
			
			if (flags & TEXT2_IMPLICIT_X)
				x += glyph->width;
		}
	}

	ui_update_screen(clipx,clipy,clipcx,clipcy);
}

void ui_desktop_save(uint32 offset, int x, int y, int cx, int cy)
{
	GR_PIXELVAL *pixels;
	GR_DRAW_ID image;
	uint8 bpp = depth / 8;
ENTER;
	offset *= bpp;

	GrSetGCMode(gc, GR_MODE_SET);

	image = GrNewPixmap(cx, cy, NULL);
	pixels = (GR_PIXELVAL*)malloc(cx * cy * depth);
	GrCopyArea(image, gc, 0, 0, cx, cy, bstore, x, y, MWROP_COPY);
	GrReadArea(image, 0, 0, cx, cy, pixels);

	cache_put_desktop(offset, cx, cy, cx*bpp, bpp, (uint8*)pixels);
	free((char*)pixels);
	GrDestroyWindow(image);
LEAVE;
}

void ui_desktop_restore(uint32 offset, int x, int y, int cx, int cy)
{
	GR_PIXELVAL *pixels;
	GR_DRAW_ID image;
	uint8 bpp = depth / 8;
ENTER;
	offset *= bpp;

	pixels = (GR_PIXELVAL*)cache_get_desktop(offset, cx, cy, bpp);
	if (pixels == NULL)
		return ;

	GrSetGCMode(gc, GR_MODE_SET);

	image = GrNewPixmap(cx, cy, NULL);
	GrArea(image, gc, 0, 0, cx, cy, (void*)pixels, MWPF_PIXELVAL);
	GrCopyArea(bstore, gc, x, y, cx, cy, image, 0, 0, MWROP_COPY);
    ui_update_screen(x, y, cx, cy);
	GrDestroyWindow(image);
LEAVE;
}

void ui_update_screen(int x, int y, int cx, int cy)
{
	if ((cx > 0) && (cy > 0))
#ifdef IPAQ
	  GrCopyArea(wnd, gc, x-dx, y-dy, cx, cy, bstore, x, y, MWROP_COPY);
#else
	  GrCopyArea(wnd, gc, x, y, cx, cy, bstore, x, y, MWROP_COPY);
#endif
}

/* placeholder function */
int ui_display_dialog(const char *msg, const char *title, int type)
{
	return 0;
}

