/*
   rdesktop: A Remote Desktop Protocol client.
   Entrypoint and utility functions
   Copyright (C) Matthew Chapman 1999-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>	/* malloc realloc free */
#include <string.h>	/* strncpy strcasecmp */
#include <unistd.h>	/* read close getuid getgid getpid getppid gethostname */
#include <netdb.h>  	/* In order to use gethostbyname() */
#include <fcntl.h>	/* open */
#include <pwd.h>	/* getpwuid */
#include <sys/stat.h>	/* stat */
#include <sys/time.h>	/* gettimeofday */
#include <sys/times.h>	/* times */
#include <errno.h>
#include "rdesktop.h"
#include "doskbcodes.h"	/* dosKeybCodes[] struct */

char username[64];
char hostname[16];
#ifdef IPAQ
#include <nano-X.h>
GR_WINDOW_ID baseWindow;
#endif
int width = 800;
int height = 600;
int keylayout = 0x409;
int backing_store = BS_AUTO;
int desktopsize_percent;
BOOL fullscreen = False;
BOOL motion = True;
BOOL orders = True;
BOOL licence = True;
int rdp_state = NOT_CONNECTED;

/* Display usage information */
static void usage(char *program)
{
	STATUS("Usage: %s [options] server\n", program);
#ifdef IPAQ
	STATUS("   -x: parent window id (root window if not specified)\n");
#endif
	STATUS("   -u: user name\n");
	STATUS("   -d: domain\n");
	STATUS("   -s: shell\n");
	STATUS("   -c: working directory\n");
	STATUS("   -p: password (for autologon, specify '-' to read password from stdin)\n");
	STATUS("   -n: client hostname\n");
	STATUS("   -w: desktop width\n");
	STATUS("   -h: desktop height\n");
	STATUS("   -P: desktop size in percentage of current screen\n");
	STATUS("   -F: full-screen mode (overrides other size switches)\n");
	STATUS("   -k: keyboard layout (us/sv/gr/etc. or hex code prefixed by '0')\n");
	STATUS("   -T: select the backing-store method (a|x|p)\n");
	STATUS("   -b: force bitmap updates\n");
	STATUS("   -m: do not send motion events\n");
	STATUS("   -l: do not request licence\n\n");
}

/* find_keyb_code: Maps alphanumeric keyboard code to
		   hexidecimal keyboard code. */
unsigned int find_keyb_code (const char* string) {

    int i = 0;
    while (dosKeybCodes[i].key != 0) {
	if (strcasecmp (string, dosKeybCodes[i].key) == 0)
	    return dosKeybCodes[i].code;
	++i;
    }

    printf("Bad syntax: \"%s\" wasn't found in keytable, valid values are:\n", string);
    while (i--)
	printf("\t%s\t-- (0x%x)\n", dosKeybCodes[i].key, dosKeybCodes[i].code);

    return 0;
}

/* Client program */
int main(int argc, char *argv[])
{
    /* struct passwd *pw;   not used anymore  -Vince */
    char *server;
    uint32 flags;
    char domain[16];
    char password[16];
    char shell[32];
    char directory[32];
    char title[32];
    int c;

    STATUS("RWin: RDP Client For Linux.\n");
    STATUS("Version "VERSION". Copyright (C) 2000-2001 Tuxia Labs.\n\n");
    STATUS("Based on rdesktop: A Remote Desktop Protocol client.\n");
    STATUS("Version 1.0.0 Copyright (C) 1999-2000 Matt Chapman.\n");

    flags = RDP_LOGON_NORMAL;
    domain[0] = password[0] = shell[0] = directory[0] = 0;

    while ((c = getopt(argc, argv, "x:u:d:s:c:p:n:w:h:k:T:P:bmlF?")) != -1) {
	switch (c) {
#ifdef IPAQ
	    case 'x':
		baseWindow = atol(optarg);
		break;
#endif
	    case 'u':
		strncpy(username, optarg, sizeof(username));
		break;

	    case 'd':
		strncpy(domain, optarg, sizeof(domain));
		break;

	    case 'p':
		flags |= RDP_LOGON_AUTO;
		if (strcmp (optarg, "-") != 0) {
		    strncpy (password, optarg, sizeof (password));
		}
		else {
		    /* Read password from terminal or stdin */
		    if (isatty (fileno (stdin))) {
		    	char *pw;
			pw = getpass ("Password: ");
			strncpy (password, pw, sizeof (password));
		    }
		    else {
		    	if (fgets ((char *)&password, sizeof (password), stdin) == NULL) {
			    ERROR ("fgets\n");
			    exit (1);
			}
		    }
		    
		    if (password[strlen(password) - 1] == '\n')
		    	password[strlen(password) - 1] = '\0';	/* chop newline */
		}
		break;

	    case 's':
		strncpy(shell, optarg, sizeof(shell));
		break;

	    case 'c':
		strncpy(directory, optarg, sizeof(directory));
		break;

	    case 'n':
		strncpy(hostname, optarg, sizeof(hostname));
		break;

	    case 'w':
		width = strtol(optarg, NULL, 10);
		break;

	    case 'h':
		height = strtol(optarg, NULL, 10);
		break;

	    case 'k':
		if (optarg[0] == '0')
		   keylayout = strtol(optarg, NULL, 16);
		else if ((keylayout = find_keyb_code (optarg)) == 0)
		   return 0;
		break;

	    case 'm':
		motion = False;
		break;

	    case 'b':
		orders = False;
		break;

	    case 'l':
		licence = False;
		break;

	    case 'T':
		switch (optarg[0]) {
		    case 'a':
				backing_store = BS_AUTO;
				break;
		    case 'x':
				backing_store = BS_XWIN;
				break;
		    case 'p':
				backing_store = BS_PIXMAP;
				break;
		    default:
				ERROR("-T: invalid parameter %c. USe 'x' to let X handle backing stores, 'p' for pixmap backing or 'a' to autodetect and use a pixmap if necessary.\n", optarg[0]);
				exit(1);
		}
		break;

		case 'P':
			desktopsize_percent = strtol(optarg, NULL, 10);
			break;

		case 'F':
			fullscreen = True;
			break;
			
	    case '?':
	    default:
		usage(argv[0]);
		return 1;
	}
    }

    if (argc - optind < 1) {
	usage(argv[0]);
	return 1;
    }

    server = argv[optind];

/*  commented because this ends up showing as root on thin clients
       -Vince   14 Jun 2001

    if (username[0] == 0) {
	pw = getpwuid(getuid());
	if ((pw == NULL) || (pw->pw_name == NULL)) {
	    STATUS("Could not determine user name.\n");
	    return 1;
	}
	strncpy(username, pw->pw_name, sizeof(username));
    }
*/
    if (hostname[0] == 0) {
	if (gethostname(hostname, sizeof(hostname)) == -1) {
	    /* Try and determine hostname by using gethostbyname() */
	    char *host = getenv("HOSTNAME");
	    struct hostent *h = gethostbyname(host);

	    if (h == NULL) {
		STATUS("Could not determine host name.\n");
    		return 1;
      	    }
	    sprintf(hostname, "%d", (unsigned int)(h->h_addr));
	}
    }

    strcpy(title, "RWin - ");
    strncat(title, server, sizeof(title));

    if (ui_create_window(title)) {
    	if (!rdp_connect(server, flags, domain, password, shell, directory)) {
			char *errormsg = (char *)malloc(500);
			snprintf(errormsg, 500, "Error connecting to server: %s", strerror(errno));
			ui_display_dialog(errormsg, "RWin", 0);
			return 1;
		}
	    STATUS("Connection successful.\n");
	    rdp_state = CONNECTED;
		rdp_main_loop();
		ui_destroy_window();
    }

    rdp_disconnect();
    return 0;
}

/* Generate a 32-byte random for the secure transport code. */
void generate_random(uint8 *random)
{
	struct stat st;
	uint32 *r = (uint32 *)random;
	int fd;

	/* If we have a kernel random device, use it. */
	if ((fd = open("/dev/urandom", O_RDONLY)) != -1)
	{
		read(fd, random, 32);
		close(fd);
		return;
	}

	/* Otherwise use whatever entropy we can gather - ideas welcome. */
	r[0] = (getpid()) | (getppid() << 16);
	r[1] = (getuid()) | (getgid() << 16);
	r[2] = times(NULL); /* system uptime (clocks) */
	gettimeofday((struct timeval *)&r[3], NULL); /* sec and usec */
	stat("/tmp", &st);
	r[5] = st.st_atime;
	r[6] = st.st_mtime;
	r[7] = st.st_ctime;
}

#ifdef DEBUG_LEAK

/* malloc; exit if out of memory */
void *xmalloc2(int l, char *s, int size)
{
	void *mem = malloc(size);
	if (mem == NULL)
	{
		ERROR("xmalloc %d\n", size);
		exit(1);
	}
	printf("A:%i:%s:%08X:%04i\n", l, s, mem, size);
	return mem;
}

/* realloc; exit if out of memory */
void *xrealloc2(int l, char *s, void *oldmem, int size)
{
	void *mem = realloc(oldmem, size);
	if (mem == NULL)
	{
		ERROR("xrealloc %d\n", size);
		exit(1);
	}
	printf("R:%i:%s:%08X:%08X:%04i\n", l, s, oldmem, mem, size);
	return mem;
}

/* free */
void xfree2(int l, char *s, void *mem)
{
	printf("F:%i:%s:%08X\n", l, s, mem);
	free(mem);
}

#else

/* malloc; exit if out of memory */
void *xmalloc(int size)
{
	void *mem = malloc(size);
	if (mem == NULL)
	{
		ERROR("xmalloc %d\n", size);
		exit(1);
	}
	return mem;
}

/* realloc; exit if out of memory */
void *xrealloc(void *oldmem, int size)
{
	void *mem = realloc(oldmem, size);
	if (mem == NULL)
	{
		ERROR("xrealloc %d\n", size);
		exit(1);
	}
	return mem;
}

/* free */
void xfree(void *mem)
{
	free(mem);
}

#endif

/* Produce a hex dump */
void hexdump(unsigned char *p, unsigned int len)
{
	unsigned char *line = p;
	unsigned int thisline, offset = 0;
	int i;

	while (offset < len)
	{
		STATUS("%04x ", offset);
		thisline = len - offset;
		if (thisline > 16)
			thisline = 16;

		for (i = 0; i < thisline; i++)
			STATUS("%02x ", line[i])

		for (; i < 16; i++)
			STATUS("   ");

		for (i = 0; i < thisline; i++)
			STATUS("%c", (line[i] >= 0x20 && line[i] < 0x7f) ? line[i] : '.');

		STATUS("\n");
		offset += thisline;
		line += thisline;
	}
}
